import pytest


def pytest_addoption(parser):
    parser.addoption("--url", action="store", default="http://localhost:3000/")
    parser.addoption("--database-uri", action="store")
    parser.addoption("--reading-time", action="store", type=int, default=5)


@pytest.fixture(scope="session")
def url(pytestconfig):
    return pytestconfig.getoption("url")


@pytest.fixture(scope="session")
def database_uri(pytestconfig):
    return pytestconfig.getoption("database_uri")


@pytest.fixture(scope="session")
def reading_time(pytestconfig):
    return pytestconfig.getoption("reading_time")



# Copyright 2019-2023 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import functools
import logging
import warnings


def _deprecated_func(msg):
    def dec(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(msg)
            return func(*args, **kwargs)

        return wrapper

    return dec


class DeprecateAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        logging.warning(f"Argument {self.option_strings} is deprecated and is ignored.")
        delattr(namespace, self.dest)


def mark_deprecated_help_strings(parser, prefix="DEPRECATED"):
    for action in parser._actions:
        if isinstance(action, DeprecateAction):
            h = action.help
            if h is None:
                action.help = prefix
            else:
                action.help = prefix + ": " + h



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime as dt
import re
from pathlib import Path

from flask_login import UserMixin
from itsdangerous import URLSafeTimedSerializer
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import relationship
from sqlalchemy.orm import validates
from werkzeug.security import check_password_hash
from werkzeug.security import generate_password_hash

import asreview.utils as utils
from asreview.webapp import DB

PASSWORD_REGEX = (
    r"^(?=.*[a-z])(?=.*[A-Z])(?=.*\d).{8,}$"  # noqa
)
EMAIL_REGEX = r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b"


class User(UserMixin, DB.Model):
    """The User model for user accounts."""

    __tablename__ = "users"
    id = Column(Integer, primary_key=True)
    identifier = Column(String(100), nullable=False, unique=True)
    origin = Column(String(100), nullable=False)
    email = Column(String(100), unique=True)
    name = Column(String(100))
    affiliation = Column(String(100))
    hashed_password = Column(String(500))
    confirmed = Column(Boolean)
    public = Column(Boolean)
    token = Column(String(150))
    token_created_at = Column(DateTime)

    projects = relationship("Project", back_populates="owner", cascade="all, delete")

    involved_in = relationship(
        "Project", secondary="collaborations", back_populates="collaborators"
    )

    pending_invitations = relationship(
        "Project",
        secondary="collaboration_invitations",
        back_populates="pending_invitations",
    )

    @validates("identifier")
    def validate_identifier(self, _key, identifier):
        if not bool(identifier):
            raise ValueError("Identifier is required")
        return identifier

    @validates("origin")
    def validate_origin(self, _key, origin):
        if not bool(origin):
            raise ValueError("Origin is required")
        return origin

    @validates("name")
    def validate_name(self, _key, name):
        if not bool(name):
            raise ValueError("Name is required")
        elif len(name) < 3:
            raise ValueError("Name must contain more than 2 characters")
        return name

    @validates("email")
    def validate_email(self, key, email):
        if key == "email" and self.origin == "asreview":
            if bool(email) is False:
                raise ValueError("Email is required when origin is 'asreview'")
            elif not User.valid_email(email):
                raise ValueError(f"Email address '{email}' is not valid")
        return email

    def __init__(
        self,
        identifier,
        origin="asreview",
        email=None,
        name=None,
        affiliation=None,
        password=None,
        confirmed=False,
        public=True,
    ):
        self.identifier = identifier
        self.origin = origin
        self.email = email
        self.name = name
        self.affiliation = affiliation
        if self.origin == "asreview":
            self.hashed_password = User.create_password_hash(password)
        self.confirmed = confirmed
        self.public = public

    def update_profile(self, email, name, affiliation,
                       old_password=None, new_password=None, public=True):

        # if there is a request to update the password, and the origin
        # is correct
        if bool(old_password) and bool(new_password) and \
                self.origin == "asreview":
            # verify the old password
            if not self.verify_password(old_password):
                raise ValueError("Provided old password is incorrect.")
            else:
                # old password is verified. The following line will raise
                # a ValueError if the new password is wrong
                self.hashed_password = User.create_password_hash(new_password)

        self.email = email
        self.name = name
        self.affiliation = affiliation
        self.public = public

        return self

    def reset_password(self, new_password):
        if self.origin == "asreview":
            self.hashed_password = User.create_password_hash(new_password)
        # reset token
        self.token = None
        self.token_created_at = None
        return self

    def set_token_data(self, secret, salt):
        """Set token data (used in email verification after
        init, and for forgot-password"""
        token, token_created_at = User.generate_token_data(secret, salt, self.email)
        self.token = token
        self.token_created_at = token_created_at
        return self

    def verify_password(self, password):
        """Verify password"""
        if bool(self.hashed_password):
            return check_password_hash(self.hashed_password, password)
        else:
            return False

    def get_name(self):
        """Get name-ish thing from user account"""
        name = self.name or self.email
        return name

    def summarize(self):
        """Summarize user account in frontend data packet"""
        return {"id": self.id, "name": self.get_name(), "email": self.email}

    def confirm_user(self):
        """This function confirms a user by setting the confirmed
        field to True and removes the token data"""
        self.confirmed = True
        self.token = None
        self.token_created_at = None
        return self

    def token_valid(self, provided_token, max_hours=24):
        """Checks whether provided token is correct and still valid"""
        # there must be a token and a timestamp
        if bool(self.token) and bool(self.token_created_at):
            # what is now
            now = dt.datetime.utcnow()
            # get time-difference in hours
            diff = (now - self.token_created_at).total_seconds()
            # return if token is correct and we are still before deadline
            return self.token == provided_token and diff <= max_hours * 3600
        else:
            return False

    @classmethod
    def generate_token_data(cls, secret, salt, email):
        """Generate a token for verification by email"""
        serializer = URLSafeTimedSerializer(secret)
        token = serializer.dumps(email, salt=salt)
        created_at = dt.datetime.utcnow()
        return token, created_at

    @classmethod
    def valid_password(cls, password):
        return re.fullmatch(PASSWORD_REGEX, password)

    @classmethod
    def valid_email(cls, email):
        return re.fullmatch(EMAIL_REGEX, email)

    @classmethod
    def create_password_hash(cls, password):
        if bool(password) and User.valid_password(password):
            return generate_password_hash(password)
        else:
            raise ValueError(f'Password "{password}" does not meet requirements.')

    def __repr__(self):
        return f"<User {self.email!r}, id: {self.id}>"


class Collaboration(DB.Model):
    __tablename__ = "collaborations"
    id = Column(Integer, primary_key=True)
    user_id = Column(
        Integer,
        ForeignKey("users.id", ondelete="cascade"),
        nullable=False
    )
    project_id = Column(
        Integer,
        ForeignKey("projects.id", ondelete="cascade"),
        nullable=False
    )
    # make sure we have unique records in this table
    __table_args__ = (
        UniqueConstraint(
            "project_id",
            "user_id",
            name="unique_records_collaboration"
        ),
    )

    def __repr__(self):
        return f"<Collaboration project:{self.project_id} user:{self.user_id}>"


class Project(DB.Model):
    """Project table"""

    __tablename__ = "projects"
    id = Column(Integer, primary_key=True)
    project_id = Column(String(250), nullable=False, unique=True)
    owner_id = Column(Integer, ForeignKey(User.id), nullable=False)
    owner = relationship("User", back_populates="projects")

    # do not delete cascade: we don't want to
    # lose users, only collaborations
    collaborators = relationship(
        "User", secondary="collaborations", back_populates="involved_in"
    )
    pending_invitations = relationship(
        "User",
        secondary="collaboration_invitations",
        back_populates="pending_invitations",
    )

    @property
    def project_path(self):
        """Returns full project path"""
        return Path(utils.asreview_path(), self.project_id)

    @property
    def folder(self):
        """Returns foldername (which is the project_id)"""
        return self.project_id

    def __repr__(self):
        return f"<Project id: {self.project_id}, owner_id: {self.owner_id}>"


class CollaborationInvitation(DB.Model):
    """Colleboration invitations"""

    __tablename__ = "collaboration_invitations"
    id = Column(Integer, primary_key=True)
    project_id = Column(
        Integer,
        ForeignKey("projects.id", ondelete="cascade"),
        nullable=False
    )
    user_id = Column(
        Integer,
        ForeignKey("users.id", ondelete="cascade"),
        nullable=False
    )
    # make sure we have unique records in this table
    __table_args__ = (
        UniqueConstraint(
            "project_id",
            "user_id",
            name="unique_records_invitations"
        ),
    )

    def __repr__(self):
        pid = self.project_id
        uid = self.user_id
        return f"<CollaborationInvitation project:{pid} user:{uid}>"


def create_database_and_tables(engine):
    """Creating database and tables with engine"""
    DB.Model.metadata.create_all(engine)



import pytest

import asreview.webapp.tests.utils.api_utils as au
from asreview.project import get_projects
from asreview.webapp import DB
from asreview.webapp.tests.utils import crud
from asreview.webapp.tests.utils.config_parser import get_user


@pytest.fixture(params=["client_auth", "client_no_auth"])
def setup(request):
    """Setup and teardown fixture that will run each test with
    an authenticated version and unauthenticated version of the
    app. In the authenticated version the fixture yields a Flask
    client, a signed-in user plus a project belongoing to this user.
    In the unauthenticated version, the fixture yields a Flask
    client, and a project."""
    # get the client
    client = request.getfixturevalue(request.param)
    # provide a project name
    project_name = "project_name"
    if request.param == "client_auth":
        # create, signup and signin users
        user1 = au.create_and_signin_user(client, 1)
        # create a project for this logged in user
        au.create_project(client, project_name)
        # receive project
        project = user1.projects[0]
    else:
        # this has to be created to match the authenticated
        # version of this fixture
        user1 = None
        # create a project
        au.create_project(client, project_name)
        # get all project
        project = get_projects()[0]
    yield client, user1, project
    if request.param == "client_auth":
        # cleanup database and asreview_path
        crud.delete_everything(DB)


@pytest.fixture(
    params=[
        "client_auth",
        "client_auth_no_creation",
        "client_auth_verified",
        "client_no_auth",
    ]
)
def setup_all_clients(request):
    """This fixture provides 4 different Flask client (authenticated
    and unauthenticated) for every test that uses it."""
    client = request.getfixturevalue(request.param)
    yield client


@pytest.fixture()
def setup_auth(client_auth):
    """This fixtures yields a Flask client for an authenticated
    app, 3 user accounts (first user is signed in) and a project
    belonging to the first user."""
    # create, signup and signin users
    user1 = au.create_and_signin_user(client_auth, 1)
    user2 = get_user(2)
    user3 = get_user(3)
    # signup user 2
    au.signup_user(client_auth, user2)
    au.signup_user(client_auth, user3)
    # get users 2 and 3 from DB
    user2 = crud.get_user_by_identifier(user2.identifier)
    user3 = crud.get_user_by_identifier(user3.identifier)
    # create a project for this logged in user
    project_name = "project_name"
    au.create_project(client_auth, project_name)
    yield client_auth, user1, user2, user3, user1.projects[0]
    # cleanup database and asreview_path
    crud.delete_everything(DB)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import logging
import os
import socket
import time
import webbrowser
from pathlib import Path
from threading import Timer

import requests
import waitress

from asreview import __version__
from asreview._deprecated import DeprecateAction
from asreview._deprecated import mark_deprecated_help_strings
from asreview.project import ASReviewProject
from asreview.project import get_project_path
from asreview.project import get_projects
from asreview.webapp.app import create_app

try:
    from gevent.pywsgi import WSGIServer
except ImportError:
    WSGIServer = None

# Host name
HOST_NAME = os.getenv("ASREVIEW_LAB_HOST", "localhost")
PORT_NUMBER = os.getenv("ASREVIEW_LAB_PORT", 5000)


def _deprecated_dev_mode():
    if os.environ.get("FLASK_DEBUG", "") == "1":
        print(
            "\n\n\n!IMPORTANT!\n\n"
            "asreview lab development mode is deprecated, see:\n"
            "https://github.com/J535D165/asreview/blob/master/DEVELOPMENT.md"
            "\n\n\n"
        )
        exit(1)


def _check_port_in_use(host, port):
    logging.info(f"Checking if host and port are available :: {host}:{port}")
    host = host.replace("https://", "").replace("http://", "")
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


def _open_browser(start_url):
    Timer(1, lambda: webbrowser.open_new(start_url)).start()

    print("\n\n\n\nIf your browser doesn't open. " f"Navigate to {start_url}\n\n\n\n")


def _check_for_update():
    """Check if there is an update available."""

    try:
        r = requests.get("https://pypi.org/pypi/asreview/json")
        r.raise_for_status()
        latest_version = r.json()["info"]["version"]
        if latest_version != __version__ and "+" not in __version__:
            print(
                "\n\n\n"
                f"ASReview LAB version {latest_version} is available. "
                "Please update using:\n"
                "pip install --upgrade asreview"
                "\n\n\n"
            )

            time.sleep(5)
    except Exception:
        print("Could not check for updates.")


def lab_entry_point(argv):
    """Entry point for the ASReview LAB webapp.

    This function is called when the `asreview lab` command is used.

    Arguments
    ---------
    argv: list
        Command line arguments.

    Examples
    --------
    >>> lab_entry_point(["--port", "5000"])
    Serving ASReview LAB at http://localhost:5000/

    Two examples of how to set the secret key for secure sessions:
    >>> ASREVIEW_LAB_SECRET_KEY="my-secret" asreview lab
    >>> asreview lab --secret-key "my-secret"
    """
    # check deprecated dev mode
    _deprecated_dev_mode()

    parser = _lab_parser()
    mark_deprecated_help_strings(parser)
    args = parser.parse_args(argv)

    # check for update
    if not args.skip_update_check:
        _check_for_update()

    app = create_app(config_path=args.config_path)

    # override config with command line arguments
    if args.secret_key:
        app.config["SECRET_KEY"] = args.secret_key

    if args.salt:
        app.config["SALT"] = args.salt

    # by default, the application is authenticated but lab is not
    app.config["LOGIN_DISABLED"] = app.config.get("LOGIN_DISABLED", True) \
        and args.login_disabled is not False

    # clean all projects
    # TODO@{Casper}: this needs a little bit
    # of work, we need to access all sub-folders
    if args.clean_all_projects:
        print("Cleaning all project files.")
        for project in get_projects():
            project.clean_tmp_files()
        print("Done")
        return

    # clean project by project_id
    # TODO@{Casper}: cleaning without a user context
    # is meaningless -> I think we should remove this
    # option
    if args.clean_project is not None:
        print(f"Cleaning project file '{args.clean_project}'.")
        ASReviewProject(get_project_path(args.clean_project)).clean_tmp_files()
        print("Done")
        return

    # if port is already taken find another one
    port = args.port
    original_port = port
    while _check_port_in_use(args.host, port) is True:
        old_port = port
        port = int(port) + 1
        if port - original_port >= args.port_retries:
            raise ConnectionError(
                "Could not find an available port \n"
                "to launch ASReview LAB. Last port \n"
                f"was {str(port)}"
            )
        print(f"Port {old_port} is in use.\n* Trying to start at {port}")

    # deprecated, remove in version 2.0
    if args.certfile or args.keyfile:
        print(
            "SSL/TLS is deprecated for the built-in server and will be removed in v2"
            "Use a dedicated WSGI server (e.g. gUnicorn) instead to make use of TLS. "
            "See https://flask.palletsprojects.com/en/3.0.x/deploying/\n\n"
        )

        if WSGIServer is None:
            raise ImportError(
                "The gevent package is required for SSL/TLS support. "
                "Please install gevent first."
            )

        protocol = "https://" if args.certfile and args.keyfile else "http://"
        start_url = f"{protocol}{args.host}:{port}/"

        ssl_args = {}
        if args.keyfile and args.certfile:
            ssl_args = {"keyfile": args.keyfile, "certfile": args.certfile}

        server = WSGIServer((args.host, port), app, **ssl_args)
        print(f"Serving ASReview LAB at {start_url}")

        if not args.no_browser:
            _open_browser(start_url)

        try:
            server.serve_forever()
        except KeyboardInterrupt:
            print("\n\nShutting down server\n\n")

    start_url = f"http://{args.host}:{port}/"

    print(f"Serving ASReview LAB at {start_url}")
    if not args.no_browser:
        _open_browser(start_url)

    try:
        waitress.serve(app, host=args.host, port=port, threads=6)
    except KeyboardInterrupt:
        print("\n\nShutting down server\n\n")


def _lab_parser():
    # parse arguments if available
    parser = argparse.ArgumentParser(
        prog="lab",
        description="ASReview LAB - Active learning for Systematic Reviews.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--clean-project",
        dest="clean_project",
        type=str,
        help="Safe cleanup of temporary files in project.",
    )

    parser.add_argument(
        "--clean-all-projects",
        dest="clean_all_projects",
        action="store_true",
        help="Safe cleanup of temporary files in all projects.",
    )

    parser.add_argument(
        "--ip",
        default=HOST_NAME,
        type=str,
        action=DeprecateAction,
        help="The IP address the server will listen on. Use the --host argument.",
    )

    parser.add_argument(
        "--host",
        default=HOST_NAME,
        type=str,
        help="The IP address the server will listen on.",
    )

    parser.add_argument(
        "--port",
        default=PORT_NUMBER,
        type=int,
        help="The port the server will listen on.",
    )

    parser.add_argument(
        "--enable-auth",
        dest="login_disabled",
        default=None,
        action="store_false",
        help="Enable authentication.",
    )

    parser.add_argument(
        "--secret-key",
        type=str,
        help="Secret key for authentication.",
    )

    parser.add_argument(
        "--salt",
        type=str,
        help="When using authentication, a salt code is needed for hasing passwords.",
    )

    parser.add_argument(
        "--config-path",
        "--flask-configfile",
        type=Path,
        help="Path to a TOML file containing ASReview parameters" "for authentication.",
    )

    parser.add_argument(
        "--no-browser",
        dest="no_browser",
        action="store_true",
        help="Do not open ASReview LAB in a browser after startup.",
    )

    parser.add_argument(
        "--port-retries",
        dest="port_retries",
        default=50,
        type=int,
        help="The number of additional ports to try if the"
        "specified port is not available.",
    )

    parser.add_argument(
        "--certfile",
        default="",
        type=str,
        help="The full path to an SSL/TLS certificate file. "
        "Use dedicated WSGI server (e.g. gUnicorn) instead to make use of TLS. "
        "See https://flask.palletsprojects.com/en/3.0.x/deploying/",
    )

    parser.add_argument(
        "--keyfile",
        default="",
        type=str,
        help="The full path to a private key file for usage with SSL/TLS. "
        "Use dedicated WSGI server (e.g. gUnicorn) instead to make use of TLS. "
        "See https://flask.palletsprojects.com/en/3.0.x/deploying/",
    )

    parser.add_argument(
        "--skip-update-check",
        dest="skip_update_check",
        action="store_true",
        help="Skip checking for updates.",
    )

    return parser



import json
from pathlib import Path

import pytest

from asreview.entry_points.simulate import SimulateEntryPoint
from asreview.entry_points.simulate import _simulate_parser
from asreview.project import ASReviewProject
from asreview.project import ProjectExistsError
from asreview.project import open_state

ADVANCED_DEPS = {"tensorflow": False}

try:
    import tensorflow  # noqa

    ADVANCED_DEPS["tensorflow"] = True
except ImportError:
    pass


DATA_FP = Path("tests", "demo_data", "generic_labels.csv")
DATA_FP_URL = "https://raw.githubusercontent.com/asreview/asreview/master/tests/demo_data/generic_labels.csv"  # noqa
DATA_FP_NO_ABS = Path("tests", "demo_data", "generic_labels_no_abs.csv")
DATA_FP_NO_TITLE = Path("tests", "demo_data", "generic_labels_no_title.csv")
EMBEDDING_FP = Path("tests", "demo_data", "generic.vec")
CFG_DIR = Path("tests", "cfg_files")
STATE_DIR = Path("tests", "state_files")
H5_STATE_FILE = Path(STATE_DIR, "test.h5")
JSON_STATE_FILE = Path(STATE_DIR, "test.json")


@pytest.mark.xfail(
    raises=FileNotFoundError,
    reason="File, URL, or dataset does not exist: " "'this_doesnt_exist.csv'",
)
def test_dataset_not_found(tmpdir):
    entry_point = SimulateEntryPoint()
    asreview_fp = Path(tmpdir, "project.asreview")
    argv = f"does_not.exist -s {asreview_fp}".split()
    entry_point.execute(argv)


def test_simulate_review_finished(tmpdir):
    # file path
    asreview_fp = Path(tmpdir, "test.asreview")

    # simulate entry point
    entry_point = SimulateEntryPoint()
    entry_point.execute(f"{DATA_FP} -s {asreview_fp}".split())

    Path(tmpdir, "test").mkdir(parents=True)
    project = ASReviewProject.load(asreview_fp, Path(tmpdir, "test"))

    assert project.config["reviews"][0]["status"] == "finished"


def test_prior_idx(tmpdir):
    asreview_fp = Path(tmpdir, "test.asreview")
    argv = f"{str(DATA_FP)} -s {asreview_fp} --prior_idx 1 4".split()
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    with open_state(asreview_fp) as state:
        labeling_order = state.get_order_of_labeling()
        query_strategies = state.get_query_strategies()

    assert labeling_order[0] == 1
    assert labeling_order[1] == 4
    assert all(query_strategies[:1] == "prior")
    assert all(query_strategies[2:] != "prior")


def test_n_prior_included(tmpdir):
    asreview_fp = Path(tmpdir, "test.asreview")
    argv = f"{str(DATA_FP)} -s {asreview_fp} --n_prior_included 2".split()
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    with open_state(asreview_fp) as state:
        result = state.get_dataset(["label", "query_strategy"])

    prior_included = result["label"] & (result["query_strategy"] == "prior")
    assert sum(prior_included) == 2

    Path(tmpdir, "test").mkdir(parents=True)
    project = ASReviewProject.load(asreview_fp, Path(tmpdir, "test"))

    settings_path = Path(
        project.project_path,
        "reviews",
        project.config["reviews"][0]["id"],
        "settings_metadata.json",
    )
    with open(settings_path, "r") as f:
        settings_metadata = json.load(f)

    assert settings_metadata["settings"]["n_prior_included"] == 2


def test_n_prior_excluded(tmpdir):
    asreview_fp = Path(tmpdir, "test.asreview")
    argv = f"{str(DATA_FP)} -s {asreview_fp} --n_prior_excluded 2".split()
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    with open_state(asreview_fp) as state:
        result = state.get_dataset(["label", "query_strategy"])

    prior_excluded = ~result["label"] & (result["query_strategy"] == "prior")
    assert sum(prior_excluded) == 2

    Path(tmpdir, "test").mkdir(parents=True)
    project = ASReviewProject.load(asreview_fp, Path(tmpdir, "test"))

    settings_path = Path(
        project.project_path,
        "reviews",
        project.config["reviews"][0]["id"],
        "settings_metadata.json",
    )
    with open(settings_path, "r") as f:
        settings_metadata = json.load(f)

    assert settings_metadata["settings"]["n_prior_excluded"] == 2


# TODO: Add random seed to settings.
# def test_seed(tmpdir):
#     asreview_fp = Path(tmpdir, 'test.asreview')
#     argv = f'{str(DATA_FP)} -s {asreview_fp} --seed 42'.split()
#     entry_point = SimulateEntryPoint()
#     entry_point.execute(argv)
#
#     with open(get_settings_metadata_path(asreview_fp), 'r') as f:
#         settings_metadata = json.load(f)
#
#     assert settings_metadata['random_seed'] == 42


def test_non_tf_models(tmpdir):
    models = ["logistic", "nb", "rf", "svm"]
    for model in models:
        print(model)
        asreview_fp = Path(tmpdir, f"test_{model}.asreview")
        argv = f"{str(DATA_FP)} -s {asreview_fp} -m {model}".split()
        entry_point = SimulateEntryPoint()
        entry_point.execute(argv)

        with open_state(asreview_fp) as state:
            classifiers = state.get_classifiers()
        default_n_priors = 2
        assert all(classifiers[default_n_priors:] == model)

        Path(tmpdir, f"test_{model}").mkdir(parents=True)
        project = ASReviewProject.load(asreview_fp, Path(tmpdir, f"test_{model}"))

        settings_path = Path(
            project.project_path,
            "reviews",
            project.config["reviews"][0]["id"],
            "settings_metadata.json",
        )
        with open(settings_path, "r") as f:
            settings_metadata = json.load(f)

        assert settings_metadata["settings"]["model"] == model


def test_last_probabilities(tmpdir):
    asreview_fp = Path(tmpdir, "test.asreview")
    argv = f"{str(DATA_FP)} -s {asreview_fp}".split()
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    with open_state(asreview_fp) as state:
        last_probabilities = state.get_last_probabilities()
    assert not last_probabilities.empty


def test_number_records_found(tmpdir):
    dataset = "synergy:van_de_Schoot_2018"
    asreview_fp = Path(tmpdir, "test.asreview")
    stop_if = 100
    priors = [116, 285]
    seed = 101

    argv = (
        f"{dataset} -s {asreview_fp} --stop_if {stop_if} "
        f"--prior_idx {priors[0]} {priors[1]} --seed {seed}".split()
    )
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    with open_state(asreview_fp) as s:
        assert s.get_labels().sum() == 29


def test_stop_if_min(tmpdir):
    dataset = "synergy:van_de_Schoot_2018"
    asreview_fp = Path(tmpdir, "test.asreview")
    stop_if = "min"
    priors = [116, 285]
    seed = 101

    argv = (
        f"{dataset} -s {asreview_fp} --stop_if {stop_if} "
        f"--prior_idx {priors[0]} {priors[1]} --seed {seed}".split()
    )
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    with open_state(asreview_fp) as s:
        assert s.get_labels().sum() == 38
        assert len(s.get_labels()) == 630


def test_stop_if_all(tmpdir):
    dataset = "synergy:van_de_Schoot_2018"
    asreview_fp = Path(tmpdir, "test.asreview")
    stop_if = -1
    priors = [116, 285]
    seed = 101

    argv = (
        f"{dataset} -s {asreview_fp} --stop_if {stop_if} "
        f"--prior_idx {priors[0]} {priors[1]} --seed {seed}".split()
    )
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    with open_state(asreview_fp) as s:
        assert s.get_labels().sum() == 38
        assert len(s.get_labels()) == 4544


def test_write_interval(tmpdir):
    dataset = "synergy:van_de_Schoot_2018"
    asreview_fp = Path(tmpdir, "test.asreview")
    stop_if = 100
    priors = [116, 285]
    seed = 101
    write_interval = 20

    argv = (
        f"{dataset} -s {asreview_fp} --stop_if {stop_if} "
        f"--prior_idx {priors[0]} {priors[1]} --seed {seed} "
        f"--write_interval {write_interval}".split()
    )
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    with open_state(asreview_fp) as s:
        assert s.get_labels().sum() == 29


@pytest.mark.xfail(raises=ProjectExistsError, reason="Cannot continue simulation.")
def test_project_already_exists_error(tmpdir):
    asreview_fp1 = Path(tmpdir, "test1.asreview")

    argv = (
        f"synergy:van_de_Schoot_2018 -s {asreview_fp1} --stop_if 100"
        f" --seed 535".split()
    )
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    # Simulate 100 queries in two steps of 50.
    argv = (
        f"synergy:van_de_Schoot_2018 -s {asreview_fp1} --stop_if 50"
        f" --seed 535".split()
    )
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)


@pytest.mark.skip(reason="Partial simulations are not available.")
def test_partial_simulation(tmpdir):
    dataset = "synergy:van_de_Schoot_2018"
    asreview_fp1 = Path(tmpdir, "test1.asreview")
    asreview_fp2 = Path(tmpdir, "test2.asreview")

    priors = [284, 285]
    seed = 101

    # Simulate 100 queries in one go.
    argv = (
        f"{dataset} -s {asreview_fp1} --stop_if 100 "
        f"--prior_idx {priors[0]} {priors[1]} --seed {seed}".split()
    )
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    # Simulate 100 queries in two steps of 50.
    argv = (
        f"{dataset} -s {asreview_fp2} --stop_if 50 "
        f"--prior_idx {priors[0]} {priors[1]} --seed {seed}".split()
    )
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    argv = (
        f"{dataset} -s {asreview_fp2} --stop_if 100 "
        f"--prior_idx {priors[0]} {priors[1]} --seed {seed}".split()
    )
    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    with open_state(asreview_fp1) as state:
        dataset1 = state.get_dataset()

    with open_state(asreview_fp2) as state:
        dataset2 = state.get_dataset()

    assert dataset1.shape == dataset2.shape
    # All query strategies should match.
    assert dataset1["query_strategy"].to_list() == dataset2["query_strategy"].to_list()
    # The first 50 record ids and labels should match.
    assert (
        dataset1["record_id"].iloc[:50].to_list()
        == dataset2["record_id"].iloc[:50].to_list()
    )
    assert (
        dataset1["label"].iloc[:50].to_list() == dataset2["label"].iloc[:50].to_list()
    )

    # You expect many of the same records in the second 50 records.
    # With this initial seed there are 89 in the total.
    assert (
        len(dataset1["record_id"][dataset1["record_id"].isin(dataset2["record_id"])])
        == 89
    )


@pytest.mark.skip(reason="Partial simulations are not available.")
def test_is_partial_simulation(tmpdir):
    dataset = "synergy:van_de_Schoot_2018"
    asreview_fp = Path(tmpdir, "test.asreview")

    argv = f"{dataset} -s {asreview_fp} --stop_if 50".split()
    parser = _simulate_parser()
    args = parser.parse_args(argv)

    assert not _is_partial_simulation(args)  # noqa

    entry_point = SimulateEntryPoint()
    entry_point.execute(argv)

    assert _is_partial_simulation(args)  # noqa



import pytest

from asreview.datasets import BaseDataGroup
from asreview.datasets import BaseDataSet
from asreview.datasets import DatasetManager
from asreview.datasets import NaturePublicationDataGroup


def test_group():
    group_nature = NaturePublicationDataGroup()

    assert group_nature.group_id is not None

    assert len(group_nature.datasets) == 4

    for d in group_nature.datasets:
        assert d.filepath.startswith(
            "https://raw.githubusercontent.com/asreview/paper-asreview"
        )


def test_group_to_dict():
    group_nature = NaturePublicationDataGroup()

    assert isinstance(group_nature.__dict__(), dict)


def test_group_list():
    dm = DatasetManager()

    nature_group = dm.list(
        include="benchmark-nature", raise_on_error=True, serialize=False
    )[0]

    assert len(nature_group.datasets) == 4

    for d in nature_group.datasets:
        assert d.filepath.startswith(
            "https://raw.githubusercontent.com/asreview/paper-asreview"
        )


def test_group_exclude_list():
    dm = DatasetManager()

    groups = dm.list(exclude="benchmark-nature", raise_on_error=True, serialize=False)
    assert "benchmark-nature" not in [group.group_id for group in groups]

    groups = dm.list(exclude=["benchmark-nature"], raise_on_error=True, serialize=False)
    assert "benchmark-nature" not in [group.group_id for group in groups]


def test_template_group():
    # START - use for building your plugin
    my_dataset1 = BaseDataSet(
        dataset_id="my_dataset1", filepath="http", title="My dataset"
    )

    my_dataset2 = BaseDataSet(
        dataset_id="my_dataset2",
        filepath="http",
        title="My second dataset",
        aliases=["J535"],
    )

    class TemplateDataGroup(BaseDataGroup):
        group_id = "template"
        description = "Template group"

        def __init__(self):
            super(TemplateDataGroup, self).__init__(my_dataset1, my_dataset2)

    # END

    my_group = TemplateDataGroup()

    my_group.find("my_dataset1").title == "My dataset"
    my_group.find("j535").title == "My second dataset"  # see lowercase alias


@pytest.mark.xfail(raises=TypeError)
def test_template_group_abc():
    my_dataset1 = BaseDataSet(
        dataset_id="my_dataset2",
        filepath="http",
        title="My second dataset",
        aliases=["J535"],
    )

    class TemplateDataGroup(BaseDataGroup):
        description = "Template group"

    TemplateDataGroup(my_dataset1)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "ClusterQuery",
    "MaxQuery",
    "MaxRandomQuery",
    "MaxUncertaintyQuery",
    "MixedQuery",
    "RandomQuery",
    "UncertaintyQuery",
    "get_query_class",
    "get_query_model",
    "list_query_strategies",
]

from asreview.models.query.cluster import ClusterQuery
from asreview.models.query.max import MaxQuery
from asreview.models.query.mixed import MaxRandomQuery
from asreview.models.query.mixed import MaxUncertaintyQuery
from asreview.models.query.mixed import MixedQuery
from asreview.models.query.random import RandomQuery
from asreview.models.query.uncertainty import UncertaintyQuery
from asreview.models.query.utils import get_query_class
from asreview.models.query.utils import get_query_model
from asreview.models.query.utils import list_query_strategies

"""Query strategies query records to label by the user.

There are several query strategies available. In configuration files,
parameters are found under the section ``[query_param]``.
"""



import pytest

import asreview.webapp.tests.utils.crud as crud
from asreview.webapp import DB


@pytest.fixture()
def setup_teardown(auth_app):
    """A fixture for an authenticated app that ensures tests are
    started with no users and projects."""
    assert crud.count_users() == 0
    assert crud.count_projects() == 0
    yield
    crud.delete_everything(DB)


@pytest.fixture()
def test_data(auth_app):
    """A fixture for an authenticated app, creates 3 users, first user
    has created 2 projects."""
    user1, _ = crud.create_user1_with_2_projects(DB)
    user2 = crud.create_user(DB, user=2)
    user3 = crud.create_user(DB, user=3)
    assert crud.count_projects() == 2
    assert crud.count_users() == 3
    data = {"user1": user1, "user2": user2, "user3": user3}
    yield data
    crud.delete_everything(DB)


@pytest.fixture()
def user(auth_app):
    """A fixture for an authenticated app, creates a single user."""
    assert crud.count_projects() == 0
    user = crud.create_user(DB, 1)
    assert crud.count_users() == 1
    yield user
    crud.delete_everything(DB)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["RISReader", "RISWriter"]

import io
import re
from urllib.request import urlopen

import pandas
import rispy

from asreview.io.utils import _standardize_dataframe
from asreview.utils import is_url

ASREVIEW_PARSE_RE = r"\bASReview_\w+\b"
ASREVIEW_PARSE_DICT = {
    "ASReview_relevant": {"included": 1},
    "ASReview_irrelevant": {"included": 0},
    "ASReview_not_seen": {"included": -1},
    "ASReview_prior": {"asreview_prior": 1},
    "ASReview_validate_relevant": {"asreview_label_to_validate": 1},
    "ASReview_validate_irrelevant": {"asreview_label_to_validate": 0},
    "ASReview_validate_not_seen": {"asreview_label_to_validate": -1},
}


def _parse_asreview_data_from_notes(note_list):

    # Return {} and an empty list
    if not isinstance(note_list, list):
        return {}

    # match all words that start with ASReview and end with a word boundary
    matches = re.findall(ASREVIEW_PARSE_RE, " ".join(note_list))

    if (
        ("ASReview_relevant" in matches and "ASReview_irrelevant" in matches)
        or ("ASReview_relevant" in matches and "ASReview_not_seen" in matches)
        or ("ASReview_irrelevant" in matches and "ASReview_not_seen" in matches)
    ):
        raise ValueError("Cannot have multiple labels for the same record.")

    if (
        (
            "ASReview_validate_relevant" in matches
            and "ASReview_validate_irrelevant" in matches
        )
        or (
            "ASReview_validate_relevant" in matches
            and "ASReview_validate_not_seen" in matches
        )
        or (
            "ASReview_validate_irrelevant" in matches
            and "ASReview_validate_not_seen" in matches
        )
    ):
        raise ValueError("Cannot have multiple labels to validate for the same record.")

    # get the dictionary for each match
    parsed_values = [ASREVIEW_PARSE_DICT.get(m, {}) for m in matches]
    parsed_values = {k: v for d in parsed_values for k, v in d.items()}

    return parsed_values


def _remove_asreview_data_from_notes(note_list):
    """Remove ASReview data from notes.

    Arguments
    ---------
    note_list: list
        A list of notes, coming from the Dataframe's "notes" column.

    Returns
    -------
    asreview_new_notes: list
        A list of updated notes, where ASReview data has been removed.
    """

    # Return {} and an empty list
    if not isinstance(note_list, list):
        return []

    asreview_new_notes = [re.sub(ASREVIEW_PARSE_RE, "", note) for note in note_list]
    asreview_new_notes[:] = [item for item in asreview_new_notes if item != ""]

    return asreview_new_notes


class RISReader:
    """RIS file reader."""

    read_format = [".ris", ".txt"]
    write_format = [".csv", ".tsv", ".xlsx", ".ris"]

    def _strip_zotero_p_tags(note_list):
        """Converter function for removing the XHTML <p></p> tags from Zotero export.

        Arguments
        ---------
        note_list: list
            A list of notes, coming from the Dataframe's "notes" column.

        Returns
        -------
        new_notes: list
            A list of updated notes, where XHTML <p></p> tags have been stripped.
        note_list: list
            The original note_list, when no XHTML <p></p> tags have been found.
        """
        if isinstance(note_list, list):
            new_notes = []
            for v in note_list:
                try:
                    new_notes.append(re.sub(r"^<p>|<\/p>$", "", v))
                except Exception:
                    new_notes.append(v)
            return new_notes
        else:
            return note_list

    @classmethod
    def _read_from_file(cls, fp, encoding="utf8"):
        with open(fp, "r", encoding=encoding) as bibliography_file:
            return list(rispy.load(bibliography_file, skip_unknown_tags=True))

    @classmethod
    def _read_from_url(cls, fp, encoding="utf8"):
        url_input = urlopen(fp)

        bibliography_file = io.StringIO(
            url_input.read().decode(encoding)
        )
        entries = list(
            rispy.load(bibliography_file, skip_unknown_tags=True)
        )
        bibliography_file.close()

        return entries

    @classmethod
    def read_data(cls, fp):
        """Import dataset.

        Arguments
        ---------
        fp: str, pathlib.Path
            File path to the RIS file.
        note_list: list
            A list of notes, coming from the Dataframe's "notes" column.

        Returns
        -------
        pandas.DataFrame:
            Dataframe with entries.

        Raises
        ------
        ValueError
            File with unrecognized encoding is used as input.
        """
        encodings = ["utf-8", "utf-8-sig", "ISO-8859-1"]
        entries = None
        for encoding in encodings:
            try:
                if is_url(fp):
                    entries = cls._read_from_url(fp, encoding=encoding)
                    break
                else:
                    entries = cls._read_from_file(fp, encoding=encoding)
                    break
            except UnicodeDecodeError:
                continue
            except Exception as e:
                raise ValueError(f"Error reading RIS file: {e}")

        if entries is None:
            raise ValueError("Cannot find proper encoding for data file")

        # Turn the entries dictionary into a Pandas dataframe
        df = pandas.DataFrame(entries)

        # Check if "notes" column is present
        if "notes" in df:
            # Strip Zotero XHTML <p> tags on "notes"
            df["notes"] = df["notes"].apply(cls._strip_zotero_p_tags)

            # strip ASReview data from notes
            df = pandas.concat(
                [
                    df,
                    pandas.DataFrame(
                        df["notes"].apply(_parse_asreview_data_from_notes).tolist(),
                    ),
                ],
                axis=1,
            )
            df["notes"] = df["notes"].apply(_remove_asreview_data_from_notes)

            # Return the standardised dataframe with label and notes separated
            return _standardize_dataframe(df)
        else:
            # Return the standardised dataframe
            return _standardize_dataframe(df)


class RISWriter:
    """RIS file writer."""

    name = "ris"
    label = "RIS"
    caution = "Available only if you imported a RIS file when creating the project"
    write_format = ".ris"

    @classmethod
    def write_data(cls, df, fp, labels=None, ranking=None):
        """Export dataset.

        Arguments
        ---------
        df: pandas.Dataframe
            Dataframe of all available record data.
        fp: str, pathlib.Path
            File path to the RIS file, if exists.
        labels: list, numpy.ndarray
            Current labels will be overwritten by these labels
            (including unlabelled). No effect if labels is None.
        ranking: list
            Reorder the dataframe according to these (internal) indices.
            Default ordering if ranking is None.

        Returns
        -------
        RIS file
            Dataframe of all available record data.
        """

        # Turn pandas DataFrame into records (list of dictionaries) for rispy
        records = df.to_dict("records")

        # Create an array for storing modified records
        records_new = []

        # Iterate over all available records
        for rec in records:

            def _notnull(v):
                if isinstance(v, list):
                    return False
                return pandas.notnull(v)

            # Remove all nan values
            rec_copy = {k: v for k, v in rec.items() if _notnull(v)}

            for m in ["authors", "keywords", "notes"]:  # AU, KW, N1
                # Check the "authors" - AU
                try:
                    rec_copy[m] = eval(rec_copy[m])
                except Exception:
                    rec_copy[m] = []

            if "included" not in rec_copy:
                rec_copy["included"] = -1

            # write the notes with ASReview data
            for k, v in ASREVIEW_PARSE_DICT.items():
                for k_df, v_df in v.items():
                    if k_df in rec_copy:
                        if rec_copy[k_df] == v_df:
                            rec_copy["notes"].insert(0, k)

            # Append the deepcopied and updated record to a new array
            records_new.append(rec_copy)

        # From buffered dataframe
        if fp is None:
            # Write the whole content to buffer
            return rispy.dumps(records_new)

        # From IO dataframe
        else:
            # Write the whole content to a file
            with open(fp, "w", encoding="utf8") as fp:
                rispy.dump(records_new, fp)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["BaseEntryPoint"]

from abc import ABC
from abc import abstractclassmethod


class BaseEntryPoint(ABC):
    """Base class for defining entry points."""

    @abstractclassmethod
    def execute(self, argv):
        """Perform the functionality of the entry point.

        Arguments
        ---------
        argv: list
            Argument list, with the entry point and program removed.
            For example, if `asreview plot X` is executed, then argv == ['X'].
        """
        raise NotImplementedError
